//
//  ViewController.swift
//  wsq_demo
//
//  Created by Cognaxon on 8/3/20.
//  Copyright © 2020 Cognaxon. All rights reserved.
//



/*
 In order to add "WSQ Image Library" to your Xcode project you need to add following script to your project:
 Build Phases -> Add New Build Phase -> New Run Script Phase
 =======================================
 export DYLIB=libWSQ_library64.so
 mkdir -p  "$TARGET_BUILD_DIR/$TARGET_NAME.app/Contents/Frameworks"
 cp -f "$SRCROOT/$DYLIB" "$TARGET_BUILD_DIR/$TARGET_NAME.app/Contents/Frameworks"
 /usr/bin/codesign --force --sign - --timestamp=none "$TARGET_BUILD_DIR/$TARGET_NAME.app/Contents/Frameworks/$DYLIB"
 install_name_tool -change $DYLIB @loader_path/../Frameworks/$DYLIB "$TARGET_BUILD_DIR/$TARGET_NAME.app/Contents/MacOS/$PRODUCT_NAME"
 =======================================
 */


// Turn Sandboxing ON to avoid runtime errors about missing iCloud support



import Cocoa




let wsq_library_handle = dlopen("libWSQ_library64.so", RTLD_LAZY)
let GenerateSerialNumber_dlsym = dlsym(wsq_library_handle, "GenerateSerialNumber")
let UnlockWSQLibrary_dlsym = dlsym(wsq_library_handle, "UnlockWSQLibrary")
let ReadImageFromFile_dlsym = dlsym(wsq_library_handle, "ReadImageFromFile")
let SaveImageToFile_dlsym = dlsym(wsq_library_handle, "SaveImageToFile")
let WSQ_decode_stream_dlsym = dlsym(wsq_library_handle, "WSQ_decode_stream")
let WSQ_encode_stream_dlsym = dlsym(wsq_library_handle, "WSQ_encode_stream")


typealias GenerateSerialNumber_alias = @convention(c) () -> UnsafePointer<CChar>
let GenerateSerialNumber = unsafeBitCast(GenerateSerialNumber_dlsym, GenerateSerialNumber_alias.self)

typealias UnlockWSQLibrary_alias = @convention(c) (UnsafePointer<CChar>) -> CInt
let UnlockWSQLibrary = unsafeBitCast(UnlockWSQLibrary_dlsym, UnlockWSQLibrary_alias.self)

typealias ReadImageFromFile_alias = @convention(c) (UnsafePointer<CChar>, UnsafeMutablePointer<CInt>, UnsafeMutablePointer<CInt>, UnsafeMutablePointer<UnsafeMutablePointer<CUnsignedChar>>) -> CInt
let ReadImageFromFile = unsafeBitCast(ReadImageFromFile_dlsym, ReadImageFromFile_alias.self)

typealias SaveImageToFile_alias = @convention(c) (UnsafePointer<CChar>, CInt, CInt, CInt, UnsafeMutablePointer<CUnsignedChar>) -> CInt
let SaveImageToFile = unsafeBitCast(SaveImageToFile_dlsym, SaveImageToFile_alias.self)

typealias WSQ_decode_stream_alias = @convention(c) (UnsafePointer<CUnsignedChar>, CInt, UnsafeMutablePointer<UnsafeMutablePointer<CUnsignedChar>>, UnsafeMutablePointer<CInt>, UnsafeMutablePointer<CInt>, UnsafeMutablePointer<CInt>, UnsafeMutablePointer<UnsafeMutablePointer<CChar>>) -> CInt
let WSQ_decode_stream = unsafeBitCast(WSQ_decode_stream_dlsym, WSQ_decode_stream_alias.self)

typealias WSQ_encode_stream_alias = @convention(c) (UnsafePointer<CUnsignedChar>, CInt, CInt, CDouble, CInt, UnsafePointer<CChar>, UnsafeMutablePointer<UnsafeMutablePointer<CUnsignedChar>>, UnsafeMutablePointer<CInt>) -> CInt
let WSQ_encode_stream = unsafeBitCast(WSQ_encode_stream_dlsym, WSQ_encode_stream_alias.self)




//var imageData : UnsafePointer<CChar>! = nil
//var imageData : UnsafeMutablePointer<UnsafeMutablePointer<CChar>>! = nil
//var imageData : UnsafeMutablePointer<CChar>! = nil;
//var imageData:UnsafeMutablePointer<CChar>! = nil

var ret: CInt = 0;
var width: CInt = 0;
var height: CInt = 0;
var ppi: CInt = 500;
var bitrate: CDouble = 0.75;

var comment_text = UnsafeMutablePointer<CChar>.alloc(512)

var output_data_stream = UnsafeMutablePointer<CUnsignedChar>.alloc(1000000)
                                                                                                         






extension NSScrollView {
    func scrollToCenter() {
        guard let docView = documentView else { return }
        
        var flipped: CGFloat = 1.0
        if (docView.flipped == true) {
            flipped = -1.0
        }
        
        let center = NSPoint(
            x: docView.bounds.midX - contentView.frame.width / 2,
           // y: docView.bounds.midY - (docView.isFlipped ? -1 : 1) * contentView.frame.height / 2
            y: docView.bounds.midY - flipped * contentView.frame.height / 2
        )
        docView.scrollPoint(center)
    }
}





class ViewController: NSViewController {
    
    
    @IBOutlet weak var wsq_demo_view: NSView!
    
    @IBOutlet weak var output_scroll_view: NSScrollView!
    @IBOutlet weak var output_image: NSImageView!
    @IBOutlet weak var openButton: NSButton!
    @IBOutlet weak var saveButton: NSButton!
    @IBOutlet weak var registerButton: NSButton!
    @IBOutlet weak var streamToWSQButton: NSButton!
    @IBOutlet weak var WSQToStreamButton: NSButton!
    
    

    

    override func viewDidLoad() {
        super.viewDidLoad()

        // Do any additional setup after loading the view.
        
        saveButton.enabled = false
        streamToWSQButton.enabled = false
        
        

//         comment_text = (char*) NULL; //need to initialize comment_text in order to avoid crash
         
         
         if (wsq_library_handle == nil)
         {print("libWSQ_library64.so dlopen error!\n"); return;}


         //When you're done with the library, dispose of it:
         //dlclose(wsq_library_handle);
    }

    
    override var representedObject: AnyObject? {
        didSet {
        // Update the view, if already loaded.
        }
    }

    
    
    @IBAction func OpenButtonAction(sender: NSButton) {

        
        
        let openDlg = NSOpenPanel();
        openDlg.title                   = "Choose image file";
        openDlg.allowedFileTypes        = ["wsq", "bmp", "tif", "tiff", "jpg", "jpeg", "png", "gif", "tga", "pcx", "psd", "cel", "iff"];
        
        // Enable the selection of files in the dialog.
        openDlg.canChooseFiles = true
        
        // Multiple files not allowed
        openDlg.allowsMultipleSelection = false;
        
        // Can't select a directory
        openDlg.canChooseDirectories = false

       
        
        if (openDlg.runModal() == NSModalResponseOK)
        {
            let file_name_nsurl = openDlg.URL // Pathname of the file
            
            if (file_name_nsurl != nil) {
                let file_name_nsstring = file_name_nsurl!.path
                
                
                
//                var imageData:UnsafeMutablePointer<CUnsignedChar>? = nil
                
                 var imageData = UnsafeMutablePointer<CUnsignedChar>.alloc(1000000)

                
                
                ret = file_name_nsstring!.withCString {
                    s1 in
                    return ReadImageFromFile(UnsafePointer(s1), &width, &height, &imageData)
                }//ret =  file_name_nsstring.withCString
 
                

                
//                print("image width = \(width)")
//                print("image height = \(height)")
              
                                


                
/*
 
https://stackoverflow.com/questions/31378120/convert-swift-string-into-cchar-pointer
                 //--------
                 int func2(char *s1, char *s2);
                 //--------
                 
                 let str1 = "first string argument"
                 let str2 = "second string argument"
                 
                 // Swift 2:
                 let result2 = str1.withCString { s1 in
                 str2.withCString { s2 in
                 func2(UnsafeMutablePointer(s1), UnsafeMutablePointer(s2))
                 }
                 }
                 
                 // Swift 3:
                 let result2 = str1.withCString { s1 in
                 str2.withCString { s2 in
                 func2(UnsafeMutablePointer(mutating: s1), UnsafeMutablePointer(mutating: s2))
                 }
                 }
                 //--------
                 
*/
                
              
                
                let colorImageRep:NSBitmapImageRep = NSBitmapImageRep(
                bitmapDataPlanes: nil,
                pixelsWide: Int(width),
                pixelsHigh: Int(height),
                bitsPerSample: 8,
                samplesPerPixel: 4,
                hasAlpha: true,
                isPlanar: false,
                colorSpaceName: NSCalibratedRGBColorSpace,
                bytesPerRow: Int(width * 4),
                bitsPerPixel: 32)!
                               
                let pixel_data:UnsafeMutablePointer<UInt8> = colorImageRep.bitmapData
                let number_of_pixels : Int = Int(height * width)
                
                
                for i in 0 ..< number_of_pixels {
                   let i4:Int = i*4
                   pixel_data[i4] = imageData[i] //Red
                   pixel_data[i4 + 1] = imageData[i + number_of_pixels] // Green
                   pixel_data[i4 + 2] = imageData[i + number_of_pixels*2] //Blue
                   pixel_data[i4 + 3] = 0xFF //Alpha
                }
               

                let colorNSImage:NSImage = NSImage(size: NSMakeSize(CGFloat(width), CGFloat(height)))
                colorNSImage.addRepresentation(colorImageRep)
                output_image.image = colorNSImage
                
                
                //--------------- begin [Fit image into output_scroll_view]
                output_scroll_view.documentView?.setFrameSize(NSMakeSize(CGFloat(width), CGFloat(height)))
                output_image.setFrameSize(NSMakeSize(CGFloat(width), CGFloat(height)))
                output_image.setFrameOrigin(NSPoint(x: 0.0, y: 0.0))
                
                output_scroll_view.scrollToCenter()
                //--------------- end [Fit image into output_scroll_view]

                
                saveButton.enabled = true
                
                
                imageData.dealloc(1000000)
                
                
            }//if (openDlg_url != nil) {
        }//if (openDlg.runModal() == NSModalResponseOK)
        
        
        
        
        
    }
    
    
    
   
    
    
    
    @IBAction func SaveButtonAction(sender: NSButton) {
        
        
        let saveDlg = NSSavePanel();
        saveDlg.title                = "Save image file";
        saveDlg.allowedFileTypes     = ["wsq", "bmp", "tif", "png", "jpg", "tga"];
        saveDlg.allowsOtherFileTypes = true
        saveDlg.extensionHidden    = false
        saveDlg.canCreateDirectories = true
        saveDlg.nameFieldStringValue = "Untitled.wsq"
               
        
        if (saveDlg.runModal() == NSModalResponseOK)
        {
            let file_name_nsurl = saveDlg.URL // Pathname of the file
            
            if (file_name_nsurl != nil) {
                let file_name_nsstring = file_name_nsurl!.path

                let file_name_extension = file_name_nsurl!.pathExtension?.lowercaseString

                var type:CInt = 1;
                
                if(file_name_extension == "wsq")
                {
                    type = 1; //WSQ;
                }
                else if(file_name_extension == "bmp")
                {
                    type = 2; //BMP;
                }
                else if(file_name_extension == "tif")
                {
                    type = 3; //TIF;
                }
                else if(file_name_extension == "png")
                {
                    type = 4; //PNG;
                }
                else if(file_name_extension == "jpg")
                {
                    type = 5; //JPG;
                }
                else if(file_name_extension == "rgb")
                {
                    type = 6; //RGB;
                }
                else if(file_name_extension == "tga")
                {
                    type = 7; //TGA;
                }
                
                                

                let imageData_for_saving = UnsafeMutablePointer<CUnsignedChar>.alloc(3 * Int(height * width))
                
                let colorNSImage:NSImage = output_image.image!
                //let colorImageRep: NSBitmapImageRep = NSBitmapImageRep(cgImage: colorNSImage.cgImage(forProposedRect: nil, context: nil, hints: nil)!)
                let colorImageRep: NSBitmapImageRep = NSBitmapImageRep(CGImage: colorNSImage.CGImageForProposedRect(nil, context: nil, hints: nil)!)
                let pixel_data:UnsafeMutablePointer<UInt8> = colorImageRep.bitmapData
              
                let number_of_pixels : Int = Int(height * width)
                
                for i in 0 ..< number_of_pixels {
                   let i4:Int = i*4
                   imageData_for_saving[i] = pixel_data[i4] //Red
                   imageData_for_saving[i + number_of_pixels] = pixel_data[i4 + 1] // Green
                   imageData_for_saving[i + number_of_pixels*2] = pixel_data[i4 + 2] //Blue
                }
                

                ret = file_name_nsstring!.withCString {
                    s1 in
                    return SaveImageToFile(UnsafePointer(s1), type, width, height, imageData_for_saving)
                }//ret =  file_name_nsstring.withCString {
  
                
                imageData_for_saving.dealloc(3 * Int(height * width))
                
            }//if (saveDlg_url != nil) {
        }//if (saveDlg.runModal() == NSModalResponseOK)
                        
        
    }
    
    
    
    
    @IBAction func WSQToSreamButtonAction(sender: NSButton) {

        let openDlg = NSOpenPanel();
        
        openDlg.title                   = "Choose WSQ file";
        openDlg.allowedFileTypes        = ["wsq"];
        
        // Enable the selection of files in the dialog.
        openDlg.canChooseFiles = true
        
        // Multiple files not allowed
        openDlg.allowsMultipleSelection = false;
        
        // Can't select a directory
        openDlg.canChooseDirectories = false
        
        
        
 
        openDlg.beginWithCompletionHandler { (result_dlg) -> Void in
        if result_dlg == NSModalResponseOK {
        
            let file_name_nsurl = openDlg.URL // Pathname of the file
            
            if (file_name_nsurl != nil) {
            
                
                var input_stream_length:CInt = 0;
                
                //var output_data_stream:UnsafeMutablePointer<CUnsignedChar>? = nil
                
                
                
                   let input_stream_data: NSData = NSData(contentsOfURL: file_name_nsurl!)!
                   input_stream_length = CInt(input_stream_data.length);
                    let input_data_stream = Array(UnsafeBufferPointer(start: UnsafePointer<UInt8>(input_stream_data.bytes), count: input_stream_data.length))
                
                                    
                   // THE DECODING OF WSQ STREAM INTO DECOMPRESSED STREAM
                   ret = WSQ_decode_stream(input_data_stream, input_stream_length, &output_data_stream, &width, &height, &ppi, &comment_text);
                    
                
                           
                
                 let colorImageRep:NSBitmapImageRep = NSBitmapImageRep(
                 bitmapDataPlanes: nil,
                 pixelsWide: Int(width),
                 pixelsHigh: Int(height),
                 bitsPerSample: 8,
                 samplesPerPixel: 4,
                 hasAlpha: true,
                 isPlanar: false,
                 colorSpaceName: NSCalibratedRGBColorSpace,
                 bytesPerRow: Int(width * 4),
                 bitsPerPixel: 32)!
                                
                 let pixel_data:UnsafeMutablePointer<UInt8> = colorImageRep.bitmapData
                 let number_of_pixels : Int = Int(height * width)
                 
                 
                 for i in 0 ..< number_of_pixels {
                    let i4:Int = i*4
                    pixel_data[i4] = output_data_stream[i] //Red
                    pixel_data[i4 + 1] = output_data_stream[i] // Green
                    pixel_data[i4 + 2] = output_data_stream[i] //Blue
                    pixel_data[i4 + 3] = 0xFF //Alpha
                 }
                

                 let colorNSImage:NSImage = NSImage(size: NSMakeSize(CGFloat(width), CGFloat(height)))
                 colorNSImage.addRepresentation(colorImageRep)
                 self.output_image.image = colorNSImage
                 
                 
                 //--------------- begin [Fit image into output_scroll_view]
                 self.output_scroll_view.documentView?.setFrameSize(NSMakeSize(CGFloat(width), CGFloat(height)))
                 self.output_image.setFrameSize(NSMakeSize(CGFloat(width), CGFloat(height)))
                 self.output_image.setFrameOrigin(NSPoint(x: 0.0, y: 0.0))
                 
                 self.output_scroll_view.scrollToCenter()
                 //--------------- end [Fit image into output_scroll_view]
                
                
                self.streamToWSQButton.enabled = true
                
                let alert: NSAlert = NSAlert()
                alert.messageText = "WSQ Image Library"
                alert.informativeText = "WSQ file was loaded into memory stream and decompressed into another memory stream.\nNow please press button \"stream -> WSQ\" to compress memory stream and save it to file.\nWSQ compressed byte array was converted to image for showing on screen."
                alert.alertStyle = NSAlertStyle.InformationalAlertStyle
                alert.addButtonWithTitle("OK")
                alert.runModal()
                
                
                
        }//if (file_name_nsurl != nil) {
        }//if result_dlg == NSApplication.ModalResponse.OK {
        }//openDlg.begin(completionHandler:) { (result_dlg) -> Void in
            

    }
    
    
    
    
    @IBAction func StreamToWSQButtonAction(sender: NSButton) {

        let saveDlg = NSSavePanel();
        saveDlg.title                = "Save WSQ file";
        saveDlg.allowedFileTypes     = ["wsq"];
        saveDlg.allowsOtherFileTypes = true
        saveDlg.extensionHidden    = false
        saveDlg.canCreateDirectories = true
        saveDlg.nameFieldStringValue = "Untitled.wsq"
        

                saveDlg.beginWithCompletionHandler{ (result_dlg) -> Void in
                if result_dlg == NSModalResponseOK {
                
                    let file_name_nsurl = saveDlg.URL // Pathname of the file
                    
                    if (file_name_nsurl != nil) {
                        
                        bitrate = 0.75;
                        ppi = 500;
                        
                        /*
                        "Cognaxon".withCString({ s1 in
                           comment_text = s1
                        });
                        */

                        
                        let input_data_stream2 = UnsafeMutablePointer<CUnsignedChar>.alloc(Int(width*height))
                                                
                        var row, row2, pixel_number, pixel_number2:CInt
                        
                        for y in 0 ..< height {
                          row = width * y
                          row2 = width * (height - y - 1)
                            
                        for x in 0 ..< width {
                            pixel_number = row + x
                            pixel_number2 = row2 + (width - x - 1)
                            
                            input_data_stream2[Int(pixel_number2)] = output_data_stream[Int(pixel_number)]
                            }//for x in 0 ..< width
                        }//for y in 0 ..< height
                        
                        
                        
                        var output_data_stream2 = UnsafeMutablePointer<CUnsignedChar>.alloc(1000000)
                        var output_stream2_length: CInt = 0;
                        
                        // THE ENCODING OF DECOMPRESSED STREAM INTO WSQ STREAM
                        ret = WSQ_encode_stream(input_data_stream2, width, height, bitrate, ppi, comment_text, &output_data_stream2, &output_stream2_length);
                        
                                                
                        
                        /*
                                         
                                        if(comment_text != NULL)
                                        {
                                            free(comment_text);
                                            comment_text = (char*) NULL; // this prevents crashing when opening next WSQ image
                                        }
                                         
                        */
                        
                        let output_data_stream2_data = NSData(bytes: output_data_stream2, length: Int(output_stream2_length))

                        do {
                            try output_data_stream2_data.writeToURL(file_name_nsurl!, options: NSDataWritingOptions.AtomicWrite)
                        } catch {
                            print(error)
                        }
                                               

                        
                        let alert: NSAlert = NSAlert()
                        alert.messageText = "WSQ Image Library"
                        alert.informativeText = "Memory stream was compressed into another memory stream and saved to WSQ file.\nNow please press button \"Open\", select the saved file, open it and have a look at the results of memory stream processing. The processed image is mirrored/flipped (UP<->DOWN and LEFT<->RIGHT)."
                        alert.alertStyle = NSAlertStyle.InformationalAlertStyle
                        alert.addButtonWithTitle("OK")
                        alert.runModal()
                        
                        
                }//if (file_name_nsurl != nil) {
                }//if result_dlg == NSApplication.ModalResponse.OK {
                }//saveDlg.begin(completionHandler:) { (result_dlg) -> Void in
        

    }

    
    
    @IBAction func RegisterButtonAction(sender: NSButton) {


        
        var serial_number_pointer:UnsafePointer<CChar>! = nil
        
        serial_number_pointer = GenerateSerialNumber()
        
        let serial_number_pointer_nsdata: NSData = NSData(bytes: serial_number_pointer, length: 71)
        let serial_number: String = String(data: serial_number_pointer_nsdata, encoding: NSUTF8StringEncoding)!
        
        
        print("Generated serial number is:\n \(serial_number)")
        
        
        let wsq_unlock_code = "my-unlock-code" // enter here your unlock code (authorization key)
        
        
        
   /*
        wsq_unlock_code.withCString {
            let wsq_unlock_code_pointer = UnsafePointer<CChar>($0)
            //let wsq_unlock_code_pointer = UnsafeMutablePointer<Int8>(mutating: $0)
            ret = UnlockWSQLibrary(wsq_unlock_code_pointer)
        }
 */
        
        ret =  wsq_unlock_code.withCString {
            s1 in
            UnlockWSQLibrary(UnsafePointer(s1))
        }
        
       
        
        
        
        if (ret == 1)
        {print("WSQ library successfully registered to this computer.\n");}
        else if (ret == 0)
        {print("Unlock code is invalid.\n");}
        else if (ret == 2)
        {print("Error writing license file. Please check directory write permisions.\n");}
        
        
        
        let alert: NSAlert = NSAlert()
        alert.messageText = "WSQ Image Library"
        alert.informativeText = "For registration please use program \"registerWSQ\" \n available in the same directory where WSQ Image Library (libWSQ_library64.so) is located.\nPlease visit http://www.cognaxon.com for more information."
        alert.alertStyle = NSAlertStyle.InformationalAlertStyle
        alert.addButtonWithTitle("OK")
        alert.runModal()
        
        
        
    }
    
    

}

